/**************************************************************************
*   Copyright (C) 2003 by Hideki Ikemoto , (c)2004 by 421                 *
*   ikemo@wakaba.jp                                                       *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/

#include "datmanager.h"
#include "qcp932codec.h"
#include "comment.h"
#include "access.h"
#include "../part/kitaconfig.h"
#include <qregexp.h>
#include <qsemaphore.h>

using namespace Kita;

/*---------------------------------------------------*/

/* DatInfo stores information of *.dat */

DatInfo::DatInfo( const KURL& url ) : m_url( url )
{

    QString idstr;
    KURL logurl = url;
    logurl.setProtocol( "k2ch" );
    QString rawdata = Kita::Access::getCacheData( logurl );
    m_linedat.clear();
    m_subject = QString::null;
    m_maxNum = 0;

    if ( rawdata.length() ) {
        QCp932Codec codec;
        QString linedata = codec.toUnicode( rawdata );

        QStringList lines = QStringList::split( "\n", linedata );
        for ( QStringList::iterator it = lines.begin(); it != lines.end(); ++it, ++m_maxNum ) {

            /* copy dat */
            m_linedat += ( *it );

            /* get id & copy */
            idstr = "none";
            QStringList list = QStringList::split( "<>", ( *it ), true );
            if ( list.size() == 5 ) {
                /* copied from Comment::parseDateId */
                QRegExp regexp( "(\\d\\d)/(\\d\\d)/(\\d\\d) (\\d\\d):(\\d\\d)( ID:(.*))?" );
                if ( regexp.search( list[ 2 ] ) != -1 ) idstr = regexp.cap( 7 );

                /* get subject */
                if ( m_subject == QString::null && list[ 4 ] != QString::null ) m_subject = list[ 4 ];
            }
            m_idlist += idstr;
        }
    }
}


const QString& DatInfo::getDat( int num )
{

    if ( num > m_maxNum ) return QString::null;

    if ( m_linedat.at( num - 1 ) == m_linedat.end() ) return QString::null;

    return *( m_linedat.at( num - 1 ) );
}


const QString& DatInfo::getId( int num )
{

    if ( num > m_maxNum ) return QString::null;

    if ( m_idlist.at( num - 1 ) == m_linedat.end() ) return QString::null;

    return *( m_idlist.at( num - 1 ) );
}


/*---------------------------------------------------*/

/* DatManager manages *.dat. */
/* Call
   Kita::DatManager::getHtml ,or
   Kita::DatManager::getDat
   in order to obtain data. */

DatInfoList DatManager::m_datInfo;
QSemaphore DatManager::m_semap( 1 );

/* private */
/* This function searches instance of DatInfo specified by url.
   If instance does not exist, create instance. */
DatInfo* DatManager::getDatInfo( const KURL& url )
{

    m_semap++;

    if ( url.isEmpty() ) {
        m_semap--;
        return NULL;
    }

    int i = 0;
    DatInfoList::Iterator it;

    /* search */
    if ( m_datInfo.count() ) {
        for ( it = m_datInfo.begin(); it != m_datInfo.end(); ++it, i++ ) {

            if ( url.host() == ( *it ) ->url().host() &&
                    url.path() == ( *it ) ->url().path() ) {

                /* LRU */
                if ( i ) {
                    m_datInfo.remove( it );
                    m_datInfo.prepend( ( *it ) );
                }

                m_semap--;
                return ( *it );
            }
        }
    }

    /* not found */

    /*create new DatInfo and insert it into list. */
    KURL daturl = url.protocol() + "://" + url.host() + url.path();

    DatInfo* datInfo;
    datInfo = new DatInfo( daturl );
    if ( datInfo->getMaxResNumber() == 0 ) { /* cache does not exist */
        delete( datInfo );

        m_semap--;
        return NULL;
    }

    m_datInfo.prepend( datInfo );

    /* delete the last item of list */
    if ( m_datInfo.count() > DMANAGER_MAXQUEUE ) {
        it = m_datInfo.fromLast();
        m_datInfo.remove( it );

        delete( *it );
    }

    m_semap--;
    return datInfo;
}


/* public */
void DatManager::deleteDat( const KURL& url )
{
    m_semap++;

    DatInfoList::Iterator it;

    for ( it = m_datInfo.begin(); it != m_datInfo.end(); ++it ) {
        if ( url.host() == ( *it ) ->url().host() &&
                url.path() == ( *it ) ->url().path() ) {
            m_datInfo.remove( it );
            delete( *it );

            m_semap--;
            return ;
        }
    }

    m_semap--;
}


/* public */
const QString& DatManager::getDat( const KURL& url, int num )
{
    DatInfo * datInfo = getDatInfo( url );
    if ( datInfo == NULL ) return QString::null;

    return datInfo->getDat( num );
}

/* public */
const QString& DatManager::getSubject( const KURL& url )
{
    DatInfo * datInfo = getDatInfo( url );
    if ( datInfo == NULL ) return QString::null;

    return datInfo->getSubject();
}


/* public */
int DatManager::getNumByID( const KURL& url, QString strid )
{

    DatInfo * datInfo = getDatInfo( url );
    if ( datInfo == NULL ) return 0;

    int num = 0;
    int maxNum = datInfo->getMaxResNumber();

    for ( int i = 1; i <= maxNum; i++ ) {
        if ( datInfo->getId( i ) == strid ) num ++;
    }

    return num;
}


/* public */
int DatManager::getMaxResNumber( const KURL& url )
{
    DatInfo * datInfo = getDatInfo( url );
    if ( datInfo == NULL ) return 0;

    return datInfo->getMaxResNumber();
}


/* public */
QString DatManager::getHtml( const KURL& url, int startnum, int endnum )
{

    DatInfo * datInfo = getDatInfo( url );
    if ( datInfo == NULL ) return QString::null;

    QString retstr = QString::null ;
    bool showAddr = KitaConfig::showMailAddress();

    for ( int num = startnum; num <= endnum; num++ ) {

        QString line = datInfo->getDat( num );

        if ( line != QString::null ) {

            Kita::Comment comment( line );

            /* check abone */
            if ( checkAbone( datInfo->getId( num ), ABONECHK_BY_ID ) ) continue;
            if ( checkAbone( comment.getName(), ABONECHK_BY_NAME ) ) continue;
            if ( checkAbone( comment.getBody(), ABONECHK_BY_WORD ) ) continue;

            retstr += comment.toHtml( num, showAddr );
        }
    }

    return retstr;
}


/* public */ /* This function is also called by kitadomtree. */
bool DatManager::checkAbone( const QString& str, int mode )
{

    QStringList & strlist = KitaConfig::aboneIDList();
    switch ( mode ) {
    case ABONECHK_BY_ID: break;
    case ABONECHK_BY_NAME: strlist = KitaConfig::aboneNameList(); break;
    case ABONECHK_BY_WORD: strlist = KitaConfig::aboneWordList(); break;
    default: return FALSE;
    }

    int i = -1;
    for ( QStringList::iterator it = strlist.begin();
            it != strlist.end(); ++it ) {
        i = str.find( ( *it ) );
        if ( i != -1 ) {
            return TRUE;
        }
    }

    return FALSE;
}
